////////////////////////////////////////////////////////////////
//                                         //
////////////////////////////////////////////////////////////////

unit USound;

interface

uses  Grids, ExtCtrls, Types;

type
     TChunkName = array[0..3] of char;   //   

     TChunk = record   //   
       ID             :TChunkName;  //  
       Size           :longword;    //  
       Name           :TChunkName;  //  (WAVE  ID=RIFF)
     end;

     TFmt = record     //   fmt
       ID             :TChunkName;  // "fmt "
       Size           :longword;    //   (16  PCM)
       Mode           :word;        //  (PCM=1)
       NumChannels    :word;        //  : =1, =2
       SampleRate     :longword;    //  : 8000 , 44100   ..
       ByteRate       :longword;    //  ,   
       BlockAlign     :word;        //     ,   
       BitsPerSample  :word;        //    : 8 , 16   ..
     end;

     TSound = class    //  
     private
       FM     :array of byte;                //  
       FFmt   :^TFmt;                        //    fmt
       FData  :^TChunk;                      //    data
       FStart :longword;                     //   
       FSG    :TStringGrid;                  //   
       FImg   :TImage;                       //   
       FDisp  :longword;                     //  
       function Adres(C,I:longword):longword;//  I   C
       function AudioFormat(N:word):string;  //   
       function TimeSize(N:longword):string; //  N 
       function GetNumChannel:longword;      //   
       function GetLen:longword;             //   
       function GetMaxLevel:longint;         //   
       function GetMinLevel:longint;         //   
       procedure InfoClear;                  //  
       procedure InfoAdd(S1,S2:string);      //  
       function GetData(C,I:longword):longint;     //  
       procedure SetData(C,I:longword; V:longint); //  
       procedure SetDisp(D:longword);              //   
       function ChannelRect(C:longint):TRect;      //   C
       procedure ImgClear;                   //  
       procedure ShowChannel(C:longint);     //   C
     public
       constructor Create;                   //  
       destructor Destroy; override;         //  
       procedure Clear;                      //  
       procedure Info;                       //    
       procedure Show;                       //  
       procedure Open(Name:string);          //  
       procedure Save(Name:string);          //  
       procedure Run;                        //  
       property Table:TStringGrid          read FSG     write FSG;     //   
       property Image:TImage               read FImg    write FImg;    //   
       property NumChannel:longword        read GetNumChannel;         //  
       property Len:longword               read GetLen;                //  
       property MaxLevel:longint           read GetMaxLevel;           //   
       property MinLevel:longint           read GetMinLevel;           //   
       property Disp:longword              read FDisp   write SetDisp; //  
       property Data[C,I:longword]:longint read GetData write SetData; //  I  C
     end;

var  Sound :TSound;    //  

implementation

uses  SysUtils, Graphics, Windows, Math;

////////////////////////////////////////////////////////////////
//                                      //
////////////////////////////////////////////////////////////////

function TSound.Adres(C,I:longword):longword;  //  I   C
begin
     result:=FStart+I*FFmt^.BlockAlign+
             C*(FFmt^.BlockAlign div NumChannel);
end;

function TSound.GetData(C,I:longword):longint;  //  
var  A,B     :longword;   //  , 
     UI      :^smallint;  //   int   2 
     UB1,UB2 :^byte;      //   byte
     j,N,Z   :longint;    // ,   , 
begin
     result:=0;  if FData=nil then Exit;
     A:=Adres(C,I);
     Case FFmt^.BitsPerSample of
        8: result:=FM[A];
       16: begin
             UI:=@FM[A];
             result:=UI^;
           end;
   17..32: begin
             //     B
             B:=0;  UB1:=@FM[A];  UB2:=@B;  Z:=1;
             N:=FFmt^.BlockAlign div NumChannel;
             for j:=1 to N do begin
                 UB2^:=UB1^;   //    
                 if (j=N) and ((UB2^ and $80)=$80) then begin
                    Z:=-1;  UB2^:=UB2^ and $7F;
                 end;
                 Inc(UB1);  Inc(UB2);
             end;
             //    
             B:=B shr (32-FFmt.BitsPerSample);
             result:=longint(B);  if Z<0 then result:=-result;
           end;
     end;
end;

procedure TSound.SetData(C,I:longword; V:longint);  //  
var  A,B     :longword;   //  , 
     UI      :^smallint;  //   int   2 
     UB1,UB2 :^byte;      //   byte
     j,N,Z   :longint;    // ,   , 
begin
     if FData=nil then Exit;
     A:=Adres(C,I);
     Case FFmt^.BitsPerSample of
        8: FM[A]:=byte(V);
       16: begin
             UI:=@FM[A];  UI^:=smallint(V);
           end;
   17..24: begin
             if V<0 then Z:=-1 else Z:=1;  V:=Abs(V);  B:=LongWord(V);
             B:=B shl (32-FFmt.BitsPerSample);
             N:=FFmt^.BlockAlign div NumChannel;  UB1:=@B;  UB2:=@FM[A];
             for j:=1 to N do begin
                 UB2^:=UB1^;  if (j=N) and (Z<1) then UB2^:=UB2^ or $80;
                 Inc(UB1);    Inc(UB2);
             end;
           end;
     end;
end;

procedure TSound.SetDisp(D:longword);  //   
begin
     FDisp:=D;  Show;
end;

function TSound.AudioFormat(N:word):string;  //   
begin
     //    .  RFC2361 ( )
     result:=IntToStr(N);
     Case N of
       $01:result:='PCM';
       $02:result:='ADPCM';
       $03:result:='IEEE_FLOAT';
       $04:result:='VSELP';
       $05:result:='IBM_CVSD';
       $06:result:='ALAW';
       $07:result:='MULAW';
       $08:result:='DTS';
       $09:result:='DRM';
       $10:result:='OKI_ADPCM';
       $11:result:='DVI_ADPCM';
       $12:result:='MEDIASPACE_ADPCM';
       $13:result:='SIERRA_ADPCM';
       $14:result:='G723_ADPCM';
       $15:result:='DIGISTD';
       $16:result:='DIGIFIX';
       $17:result:='DIALOGIC_OKI_ADPCM';
       $18:result:='MEDIAVISION_ADPCM';
       $19:result:='CU_CODEC';
       $20:result:='YAMAHA_ADPCM';
     end;
end;

function TSound.TimeSize(N:longword):string;  //  N 
var  H,M,S :longword;  // , , 
begin
     H:=N div 3600;
     M:=(N-H*3600) div 60;
     S:=N-H*3600-M*60;
     result:='';
     if H>0 then result:=IntToStr(H)+' . ';
     if M>0 then result:=result+IntToStr(M)+' . ';
     result:=result+IntToStr(S)+' .';
end;

function TSound.GetNumChannel:longword;  //   
begin
     if FFmt=nil then result:=2 else result:=FFmt^.NumChannels;
end;

function TSound.GetLen:longword;  //   
begin
     if (FFmt=nil) or (FData=nil)
        then result:=0
        else result:=FData^.Size div FFmt^.blockAlign;
end;

function TSound.GetMaxLevel:longint;  //   
begin
     if FFmt=nil then result:=32767 else
        if FFmt.BitsPerSample>8
           then result:=Round(Power(2,FFmt.BitsPerSample)/2-1)
           else result:=Round(Power(2,FFmt.BitsPerSample)-1);
end;

function TSound.GetMinLevel:longint;  //   
begin
     if FFmt=nil then result:=-32768 else
        if FFmt.BitsPerSample>8
           then result:=-Round(Power(2,FFmt.BitsPerSample)/2)
           else result:=0;
end;

procedure TSound.InfoClear;  //  
begin
     if FSG=nil then Exit;
     FSG.RowCount:=2;
     FSG.Cells[0,0]:='';     FSG.Cells[1,0]:='';
     FSG.Cells[0,1]:=' '; FSG.Cells[1,1]:='';
end;

procedure TSound.InfoAdd(S1,S2:string);  //  
var  L,I :longint;  // , 
begin
     if FSG=nil then Exit;
     //   
     L:=FSG.RowCount;
     for I:=FSG.RowCount-1 downto 1 do if FSG.Cells[1,I]='' then L:=I;
     if L=FSG.RowCount then FSG.RowCount:=L+1;
     //  
     FSG.Cells[0,L]:=S1;  FSG.Cells[1,L]:=S2;
end;


////////////////////////////////////////////////////////////////
//                                              //
////////////////////////////////////////////////////////////////

constructor TSound.Create;  //  
begin
     SetLength(FM,0);
     FFmt:=nil;
     FData:=nil;
     FStart:=0;
     FSG:=nil;
     FImg:=nil;
     FDisp:=0;
end;

destructor TSound.Destroy;  //  
begin
     SetLength(FM,0);
end;

procedure TSound.Clear;  //  
begin
     SetLength(FM,0);
     FFmt:=nil;  FData:=nil;  FStart:=0;
     Info;
     Show;
end;

procedure TSound.Info;  //    
var  PC :^TChunk;   //   
     A  :longword;  //   
     N  :longint;   //  
begin
     InfoClear;
     if Length(FM)=0 then PC:=nil else PC:=@FM[0];
     if (FSG=nil) or (PC=nil) then Exit;
     InfoAdd(' ',IntToStr(Length(FM)));
     InfoAdd('',PC^.ID);
     InfoAdd('',IntToStr(PC^.Size));
     InfoAdd('',PC^.Name);
     A:=12;  N:=1;
     While A<longword(Length(FM)) do begin
           PC:=@FM[A];
           InfoAdd(' '+IntToStr(N),PC^.ID);
           InfoAdd('  '+IntToStr(N),IntToStr(PC^.Size));
           if PC^.ID='fmt ' then FFmt:=@FM[A];
           if PC^.ID='data' then begin  FData:=@FM[A];  FStart:=A+8;  end;
           N:=N+1;
           A:=A+PC^.Size+8;
     end;
     if FFmt<>nil then begin
        InfoAdd(' ',AudioFormat(FFmt^.Mode));
        InfoAdd(' ',IntToStr(NumChannel));
        InfoAdd(' .',IntToStr(FFmt^.sampleRate));
        InfoAdd('  ',IntToStr(FFmt^.byteRate));
        InfoAdd('  ',IntToStr(FFmt^.blockAlign));
        InfoAdd('  ',IntToStr(FFmt^.bitsPerSample));
        if FData<>nil then begin
           InfoAdd('',IntToStr(Len));
           InfoAdd('',TimeSize(FData^.Size div FFmt^.blockAlign div FFmt^.sampleRate));
        end;
     end;
end;

function TSound.ChannelRect(C:longint):TRect;  //   C
var  H,D,N :longint;  //  , ,  
begin
     if FFmt<>nil then N:=FFmt.NumChannels else N:=2;
     H:=(FImg.ClientHeight-10) div N;
     D:=(FImg.ClientHeight-H*N) div (N+1);
     Result:=FImg.ClientRect;
     Result.Top:=FImg.ClientRect.Top+D+C*(H+D);
     Result.Bottom:=Result.Top+H;
end;

procedure TSound.ImgClear;  //  
begin
     if FImg=nil then Exit;
     FImg.Canvas.Brush.Color:=clBtnFace;
     FImg.Canvas.FillRect(FImg.ClientRect);
end;

procedure TSound.ShowChannel(C:longint);  //   C
var  R    :TRect;    //  
     L    :longint;  // 
     Xmin :longint;  //   
     Xmax :longint;  //   
     Ymin :longint;  //   
     Ymax :longint;  //   

  function Xe(X:longint):longword;  //     X
  begin
       result:=Round(R.Left+(X-Xmin)*(R.Right-R.Left)/(Xmax-Xmin));
  end;

  function Ye(Y:longint):longint;  //     Y
  begin
       result:=Round(R.Bottom+(Y-Ymin)*((R.Top-R.Bottom)/(Ymax-Ymin)));
  end;

begin
     if FImg=nil then Exit;
     R:=ChannelRect(C);
     Xmin :=Round((Len-longword(R.Right-R.Left+1))*FDisp/10000); //   
     Xmax :=Xmin+(R.Right-R.Left);                               //   
     Ymin :=MinLevel;                                            //   
     Ymax :=MaxLevel;                                            //   
     //  
     FImg.Canvas.Brush.Color:=RGB($FF,$FF,$CC);   FImg.Canvas.FillRect(R);
     FImg.Canvas.Font.Color:=clBlack;             FImg.Canvas.TextOut(R.Left+10,R.Top+10,' '+IntToStr(C));
     FImg.Canvas.Brush.Color:=clGray;             FImg.Canvas.FrameRect(R);
     //  
     FImg.Canvas.Pen.Color:=clGray;
     FImg.Canvas.MoveTo(Xe(Xmin),Ye(0));  FImg.Canvas.LineTo(Xe(Xmax),Ye(0));
     // 
     FImg.Canvas.Pen.Color:=clBlue;
     FImg.Canvas.MoveTo(Xe(XMin),Ye(Data[C,Xmin]));
     for L:=Xmin+1 to Xmax do begin
         FImg.Canvas.LineTo(Xe(L),Ye(Data[C,L]));
     end;
end;

procedure TSound.Show;  //  
var  C :longint;  // 
begin
     ImgClear;
     for C:=0 to NumChannel-1 do ShowChannel(C);
end;

procedure TSound.Open(Name:string);  //  
var  F        :file of byte;  //   
     Len,Size :longint;       //  ,  
begin
     if not FileExists(Name) then Exit;
     AssignFile(F,Name);
     Reset(F);
     Len:=FileSize(F);
     SetLength(FM,Len);
     BlockRead(F,FM[0],Len,Size);
     CloseFile(F);
     SetLength(FM,Size);
     Info;
     Show;
end;

procedure TSound.Save(Name:string);  //  
var  F :file of byte;  //   
begin
     AssignFile(F,Name);
     Rewrite(F);
     BlockWrite(F,FM[0],Length(FM));
     CloseFile(F);
end;

procedure TSound.Run;  //  
var  V   :longint;   //   
     L,C :longword;  // , 
begin
     //      "-"
     for C:=0 to NumChannel-1 do begin
         for L:=0 to (Len-1) div 2 do begin
             V:=Data[C,L];
             Data[C,L]:=Data[C,Len-1-L];
             Data[C,Len-1-L]:=V;
         end;
     end;
     //  
     Show;
end;


initialization
    Sound:=TSound.Create;  //      
finalization
    Sound.Free;            //  
end.
